
.. ifconfig:: format == 'html'

	.. toctree::
		:maxdepth: 2

		examples/index

.. toctree::
    :maxdepth: 2

    gettingstarted/index
    user-guide
    guide-index
    tutorials-index
    api-index

